@@ -2,7 +2,7 @@
 # vim: tabstop=4 shiftwidth=4 softtabstop=4
 
 # Copyright 2013 IBM Corporation
-# Copyright 2015 Lenovo
+# Copyright 2015-2017 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -35,6 +35,12 @@ class UnrecognizedCertificate(Exception):
         self.certdata = certdata
 
 
+class TemporaryError(Exception):
+    # A temporary condition that should clear, but warrants reporting to the
+    # caller
+    pass
+
+
 class InvalidParameterValue(PyghmiException):
     pass
 
@@ -48,3 +54,9 @@ class UnsupportedFunctionality(PyghmiException):
     # Indicates when functionality is requested that is not supported by
     # current endpoint
     pass
+
+
+class BypassGenericBehavior(PyghmiException):
+    # Indicates that an OEM handler wants to abort any standards based
+    # follow up
+    pass
