@@ -29,13 +29,39 @@ import pyghmi.media as media
 import pyghmi.storage as storage
 import pyghmi.util.webclient as webclient
 import random
+import re
 import socket
 import struct
-import threading
 import urllib
 import weakref
 
 
+numregex = re.compile('([0-9]+)')
+
+
+def naturalize_string(key):
+    """Analyzes string in a human way to enable natural sort
+
+    :param nodename: The node name to analyze
+    :returns: A structure that can be consumed by 'sorted'
+    """
+    return [int(text) if text.isdigit() else text.lower()
+            for text in re.split(numregex, key)]
+
+
+def natural_sort(iterable):
+    """Return a sort using natural sort if possible
+
+    :param iterable:
+    :return:
+    """
+    try:
+        return sorted(iterable, key=naturalize_string)
+    except TypeError:
+        # The natural sort attempt failed, fallback to ascii sort
+        return sorted(iterable)
+
+
 def fixup_uuid(uuidprop):
     baduuid = ''.join(uuidprop.split())
     uuidprefix = (baduuid[:8], baduuid[8:12], baduuid[12:16])
@@ -44,17 +70,11 @@ def fixup_uuid(uuidprop):
     return '-'.join(uuid).upper()
 
 
-class FileUploader(threading.Thread):
-
-    def __init__(self, webclient, url, filename, data):
-        self.wc = webclient
-        self.url = url
-        self.filename = filename
-        self.data = data
-        super(FileUploader, self).__init__()
-
-    def run(self):
-        self.rsp = self.wc.upload(self.url, self.filename, self.data)
+def fixup_str(propstr):
+    if propstr is None:
+        return ''
+    return ''.join([chr(int(c, 16)) for c in propstr.split()]).strip(
+        ' \xff\x00')
 
 
 class IMMClient(object):
@@ -128,14 +148,15 @@ class IMMClient(object):
             return None
         return cls._parse_builddate(propstr)
 
-    def get_system_configuration(self):
+    def get_system_configuration(self, hideadvanced=True):
         if not self.fwc:
             self.fwc = config.LenovoFirmwareConfig(self.ipmicmd)
         self.fwo = self.fwc.get_fw_options()
         self.fwovintage = util._monotonic_time()
         retcfg = {}
         for opt in self.fwo:
-            if self.fwo[opt]['lenovo_protect'] or self.fwo[opt]['hidden']:
+            if (hideadvanced and self.fwo[opt]['lenovo_protect'] or
+                    self.fwo[opt]['hidden']):
                 # Do not enumerate hidden settings
                 continue
             retcfg[opt] = {}
@@ -168,19 +189,26 @@ class IMMClient(object):
                 changeset[key] = {'value': changeset[key]}
             newvalue = changeset[key]['value']
             if self.fwo[key]['is_list'] and not isinstance(newvalue, list):
-                newvalues = newvalue.split(',')
+                if '=' in newvalue:
+                    # ASU set a precedent of = delimited settings
+                    # for now, honor that delimiter as well
+                    newvalues = newvalue.split('=')
+                else:
+                    newvalues = newvalue.split(',')
             else:
                 newvalues = [newvalue]
             newnewvalues = []
             for newvalue in newvalues:
+                newv = re.sub('\s+', ' ', newvalue)
                 if (self.fwo[key]['possible'] and
                         newvalue not in self.fwo[key]['possible']):
                     candlist = []
                     for candidate in self.fwo[key]['possible']:
-                        if newvalue.lower().startswith(candidate.lower()):
+                        candid = re.sub('\s+', ' ', candidate)
+                        if newv.lower().startswith(candid.lower()):
                             newvalue = candidate
                             break
-                        if candidate.lower().startswith(newvalue.lower()):
+                        if candid.lower().startswith(newv.lower()):
                             candlist.append(candidate)
                     else:
                         if len(candlist) == 1:
@@ -243,6 +271,13 @@ class IMMClient(object):
                 if 'token2_name' in rspdata and 'token2_value' in rspdata:
                     wc.set_header(rspdata['token2_name'],
                                   rspdata['token2_value'])
+                if 'token3_name' in rspdata and 'token3_value' in rspdata:
+                    self.uploadtoken = {rspdata['token3_name']:
+                                        rspdata['token3_value']}
+                else:
+                    self.uploadtoken = {}
+                wc.set_header('Referer', self.adp_referer)
+                wc.set_header('Host', self.imm)
                 return wc
 
     @property
@@ -274,31 +309,96 @@ class IMMClient(object):
         except KeyError:
             return None
 
+    def upload_media(self, filename, progress=None):
+        xid = random.randint(0, 1000000000)
+        alloc = self.wc.grab_json_response(
+            '/data/set',
+            'RP_VmAllocateLoc({0},{1},1)'.format(self.username, filename))
+        if alloc['return'] != 'Success':
+            raise Exception('Unexpected reply to allocation: ' + repr(alloc))
+        slotid = alloc['slotId']
+        uploadfields = self.uploadtoken
+        uploadfields['filePath'] = alloc['filePath']
+        uploadfields['uploadType'] = 'iframe'
+        uploadfields['available'] = alloc['available']
+        uploadfields['checksum'] = xid
+        ut = webclient.FileUploader(
+            self.wc, '/designs/imm/upload/rp_image_upload.esp', filename,
+            otherfields=uploadfields)
+        ut.start()
+        while ut.isAlive():
+            ut.join(3)
+            if progress:
+                progress({'phase': 'upload',
+                          'progress': 100 * self.wc.get_upload_progress()})
+        status = self.wc.grab_json_response(
+            '/designs/imm/upload/rp_image_upload_status.esp',
+            'filePath={0}'.format(alloc['filePath']))
+        if not status['rpImgUploadResult'].endswith('Success'):
+            raise Exception(
+                'Upload status returned unexpected data: ' + repr(alloc))
+        ups = self.wc.grab_json_response(
+            '/data/set',
+            'RP_VmUpdateSize({1}, {0})'.format(status['originalFileSize'],
+                                               slotid))
+        if ups['return'] != 'Success':
+            raise Exception('Unexpected return to update size: ' + repr(ups))
+        ups = self.wc.grab_json_response('/data/set',
+                                         'RP_VmMount({0})'.format(slotid))
+        if ups['return'] != 'Success':
+            raise Exception('Unexpected return to mount: ' + repr(ups))
+        if progress:
+            progress({'phase': 'complete'})
+
     def attach_remote_media(self, url, user, password):
         url = url.replace(':', '\:')
         params = urllib.urlencode({
             'RP_VmAllocateMountUrl({0},{1},1,,)'.format(
                 self.username, url): ''
         })
-        result = self.wc.grab_json_response('/data?set', params)
+        result = self.wc.grab_json_response('/data?set', params,
+                                            referer=self.adp_referer)
+        if not result:
+            result = self.wc.grab_json_response('/data/set', params,
+                                                referer=self.adp_referer)
         if result['return'] != 'Success':
             raise Exception(result['reason'])
         self.weblogout()
 
+    def list_media(self):
+        rt = self.wc.grab_json_response(
+            '/designs/imm/dataproviders/imm_rp_images.php',
+            referer=self.adp_referer)
+        for item in rt['items']:
+            if 'images' in item:
+                for uload in item['images']:
+                    if uload['status'] != 0:
+                        yield media.Media(uload['filename'])
+
     def detach_remote_media(self):
         mnt = self.wc.grab_json_response(
-            '/designs/imm/dataproviders/imm_rp_images.php')
+            '/designs/imm/dataproviders/imm_rp_images.php',
+            referer=self.adp_referer)
         removeurls = []
         for item in mnt['items']:
             if 'urls' in item:
                 for url in item['urls']:
                     removeurls.append(url['url'])
+            if 'images' in item:
+                for uload in item['images']:
+                    self.wc.grab_json_response(
+                        '/data/set', 'RP_RemoveFile({0}, 0)'.format(
+                            uload['slotId']))
         for url in removeurls:
             url = url.replace(':', '\:')
             params = urllib.urlencode({
                 'RP_VmAllocateUnMountUrl({0},{1},0,)'.format(
                     self.username, url): ''})
-            result = self.wc.grab_json_response('/data?set', params)
+            result = self.wc.grab_json_response('/data?set', params,
+                                                referer=self.adp_referer)
+            if not result:
+                result = self.wc.grab_json_response('/data/set', params,
+                                                    referer=self.adp_referer)
             if result['return'] != 'Success':
                 raise Exception(result['reason'])
         self.weblogout()
@@ -316,8 +416,14 @@ class IMMClient(object):
                     self.datacache['lenovo_cached_adapters'] = (
                         adapterdata, util._monotonic_time())
         if adapterdata and 'items' in adapterdata:
+            anames = {}
             for adata in adapterdata['items']:
                 aname = adata[self.ADP_NAME]
+                if aname in anames:
+                    anames[aname] += 1
+                    aname = '{0} {1}'.format(aname, anames[aname])
+                else:
+                    anames[aname] = 1
                 donenames = set([])
                 for fundata in adata[self.ADP_FUN]:
                     fdata = fundata.get('firmwares', ())
@@ -370,12 +476,12 @@ class IMMClient(object):
 
     def get_hw_inventory(self):
         hwmap = self.hardware_inventory_map()
-        for key in hwmap:
+        for key in natural_sort(hwmap):
             yield (key, hwmap[key])
 
     def get_hw_descriptions(self):
         hwmap = self.hardware_inventory_map()
-        for key in hwmap:
+        for key in natural_sort(hwmap):
             yield key
 
     def get_component_inventory(self, compname):
@@ -386,9 +492,12 @@ class IMMClient(object):
             return None
 
     def get_oem_sensor_names(self, ipmicmd):
-        if self._energymanager is None:
-            self._energymanager = energy.EnergyManager(ipmicmd)
-        return self._energymanager.supportedmeters
+        try:
+            if self._energymanager is None:
+                self._energymanager = energy.EnergyManager(ipmicmd)
+            return self._energymanager.supportedmeters
+        except pygexc.UnsupportedFunctionality:
+            return ()
 
     def get_oem_sensor_descriptions(self, ipmicmd):
         return [{'name': x, 'type': 'Energy'
@@ -425,9 +534,13 @@ class IMMClient(object):
         enclosureuuid = self.get_property('/v2/ibmc/smm/chassis/uuid')
         if enclosureuuid:
             bay = self.get_property('/v2/cmm/sp/7')
+            serial = self.get_property('/v2/ibmc/smm/chassis/sn')
+            model = self.get_property('/v2/ibmc/smm/chassis/mtm')
             hwmap['Enclosure'] = {
                 'UUID': fixup_uuid(enclosureuuid),
                 'Bay': bay,
+                'Model': fixup_str(model),
+                'Serial': fixup_str(serial),
             }
         adapterdata = self.get_cached_data('lenovo_cached_adapters')
         if not adapterdata:
@@ -441,11 +554,17 @@ class IMMClient(object):
                     self.datacache['lenovo_cached_adapters'] = (
                         adapterdata, util._monotonic_time())
         if adapterdata and 'items' in adapterdata:
+            anames = {}
             for adata in adapterdata['items']:
                 skipadapter = False
                 if not adata[self.ADP_OOB]:
                     continue
                 aname = adata[self.ADP_NAME]
+                if aname in anames:
+                    anames[aname] += 1
+                    aname = '{0} {1}'.format(aname, anames[aname])
+                else:
+                    anames[aname] = 1
                 clabel = adata[self.ADP_LABEL]
                 if clabel == 'Unknown':
                     continue
@@ -560,7 +679,12 @@ class XCCClient(IMMClient):
         super(XCCClient, self).__init__(ipmicmd)
         self.adp_referer = None
 
-    def get_webclient(self):
+    def get_description(self):
+        dsc = self.wc.grab_json_response('/DeviceDescription.json')
+        dsc = dsc[0]
+        return {'height': int(dsc['u-height']), 'slot': int(dsc['slot'])}
+
+    def get_webclient(self, login=True):
         cv = self.ipmicmd.certverify
         wc = webclient.SecureHTTPConnection(self.imm, 443, verifycallback=cv)
         try:
@@ -569,6 +693,8 @@ class XCCClient(IMMClient):
             if se.errno != errno.ECONNREFUSED:
                 raise
             return None
+        if not login:
+            return wc
         adata = json.dumps({'username': self.username,
                             'password': self.password
                             })
@@ -865,7 +991,7 @@ class XCCClient(IMMClient):
                     standalonedisks.append(
                         storage.Disk(
                             name=disk['name'], description=disk['type'],
-                            id=(cid, disk['id']),  status=disk['RAIDState'],
+                            id=(cid, disk['id']), status=disk['RAIDState'],
                             serial=disk['serialNo'], fru=disk['fruPartNo']))
         return storage.ConfigSpec(disks=standalonedisks, arrays=pools)
 
@@ -895,6 +1021,9 @@ class XCCClient(IMMClient):
             if rt['return'] in (657, 659, 656):
                 raise pygexc.InvalidParameterValue(
                     'Given location was unreachable by the XCC')
+            if rt['return'] == 32:
+                raise pygexc.InvalidParameterValue(
+                    'XCC does not have required license for operation')
             raise Exception('Unhandled return: ' + repr(rt))
         rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_mountall',
                                         '{}')
@@ -1046,9 +1175,8 @@ class XCCClient(IMMClient):
 
     def upload_media(self, filename, progress=None):
         xid = random.randint(0, 1000000000)
-        uploadthread = FileUploader(self.wc,
-                                    '/upload?X-Progress-ID={0}'.format(xid),
-                                    filename, None)
+        uploadthread = webclient.FileUploader(
+            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, None)
         uploadthread.start()
         while uploadthread.isAlive():
             uploadthread.join(3)
@@ -1112,6 +1240,16 @@ class XCCClient(IMMClient):
         if '_csrf_token' in wc.cookies:
             wc.set_header('X-XSRF-TOKEN', self.wc.cookies['_csrf_token'])
 
+    def set_hostname(self, hostname):
+        self.wc.grab_json_response('/api/dataset', {'IMM_HostName': hostname})
+        self.wc.grab_json_response('/api/dataset', {'IMM_DescName': hostname})
+        self.weblogout()
+
+    def get_hostname(self):
+        rsp = self.wc.grab_json_response('/api/dataset/sys_info')
+        self.weblogout()
+        return rsp['items'][0]['system_name']
+
     def update_firmware_backend(self, filename, data=None, progress=None,
                                 bank=None):
         self.weblogout()
@@ -1123,9 +1261,8 @@ class XCCClient(IMMClient):
         if rsv['return'] != 0:
             raise Exception('Unexpected return to reservation: ' + repr(rsv))
         xid = random.randint(0, 1000000000)
-        uploadthread = FileUploader(self.wc,
-                                    '/upload?X-Progress-ID={0}'.format(xid),
-                                    filename, data)
+        uploadthread = webclient.FileUploader(
+            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, data)
         uploadthread.start()
         uploadstate = None
         while uploadthread.isAlive():
@@ -1242,3 +1379,11 @@ class XCCClient(IMMClient):
         if bank == 'backup':
             return 'complete'
         return 'pending'
+
+    def get_health(self, summary):
+        wc = self.get_webclient(False)
+        rsp = wc.grab_json_response('/api/providers/imm_active_events')
+        if 'items' in rsp and len(rsp['items']) == 0:
+            # The XCC reports healthy, no need to interrogate
+            raise pygexc.BypassGenericBehavior()
+        # Will use the generic handling for unhealthy systems
