@@ -20,13 +20,16 @@ import json
 import pyghmi.exceptions as pygexc
 import socket
 import ssl
+import threading
 
 try:
     import Cookie
     import httplib
+    import StringIO
 except ImportError:
     import http.client as httplib
     import http.cookies as Cookie
+    import io as StringIO
 
 __author__ = 'jjohnson2'
 
@@ -39,6 +42,23 @@ BND = 'TbqbLUSn0QFjx9gxiQLtgBK4Zu6ehLqtLs4JOBS50EgxXJ2yoRMhTrmRXxO1lkoAQdZx16'
 uploadforms = {}
 
 
+class FileUploader(threading.Thread):
+
+    def __init__(self, webclient, url, filename, data=None, formname=None,
+                 otherfields=()):
+        self.wc = webclient
+        self.url = url
+        self.filename = filename
+        self.data = data
+        self.otherfields = otherfields
+        self.formname = formname
+        super(FileUploader, self).__init__()
+
+    def run(self):
+        self.rsp = self.wc.upload(self.url, self.filename, self.data,
+                                  self.formname, otherfields=self.otherfields)
+
+
 def get_upload_form(filename, data, formname, otherfields):
     if not formname:
         formname = filename
@@ -69,6 +89,7 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
                  **kwargs):
         if 'timeout' not in kwargs:
             kwargs['timeout'] = 60
+        self.broken = False
         self.thehost = host
         self.theport = port
         httplib.HTTPConnection.__init__(self, host, port, strict, **kwargs)
@@ -107,12 +128,16 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
                                                  bincert)
 
     def getresponse(self):
-        rsp = super(SecureHTTPConnection, self).getresponse()
-        for hdr in rsp.msg.headers:
-            if hdr.startswith('Set-Cookie:'):
-                c = Cookie.BaseCookie(hdr[11:])
-                for k in c:
-                    self.cookies[k] = c[k].value
+        try:
+            rsp = super(SecureHTTPConnection, self).getresponse()
+            for hdr in rsp.msg.headers:
+                if hdr.startswith('Set-Cookie:'):
+                    c = Cookie.BaseCookie(hdr[11:])
+                    for k in c:
+                        self.cookies[k] = c[k].value
+        except httplib.BadStatusLine:
+            self.broken = True
+            raise
         return rsp
 
     def grab_json_response(self, url, data=None, referer=None):
@@ -141,11 +166,15 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
         """
         if data is None:
             data = open(filename, 'rb')
-        form = get_upload_form(filename, data, formname, otherfields)
+        self._upbuffer = StringIO.StringIO(get_upload_form(filename, data,
+                                                           formname,
+                                                           otherfields))
         ulheaders = self.stdheaders.copy()
         ulheaders['Content-Type'] = 'multipart/form-data; boundary=' + BND
+        ulheaders['Content-Length'] = len(uploadforms[filename])
+        self.ulsize = len(uploadforms[filename])
         webclient = self.dupe()
-        webclient.request('POST', url, form, ulheaders)
+        webclient.request('POST', url, self._upbuffer, ulheaders)
         rsp = webclient.getresponse()
         # peer updates in progress should already have pointers,
         # subsequent transactions will cause memory to needlessly double,
@@ -159,11 +188,16 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
                             rsp.read())
         return rsp.read()
 
+    def get_upload_progress(self):
+        return float(self._upbuffer.tell()) / float(self.ulsize)
+
     def request(self, method, url, body=None, headers=None, referer=None):
         if headers is None:
             headers = self.stdheaders.copy()
         if method == 'GET' and 'Content-Type' in headers:
             del headers['Content-Type']
+        if method == 'POST' and body and 'Content-Type' not in headers:
+            headers['Content-Type'] = 'application/x-www-form-urlencoded'
         if self.cookies:
             cookies = []
             for ckey in self.cookies:
@@ -175,5 +209,9 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
                 headers['Cookie'] += '; ' + '; '.join(cookies)
         if referer:
             headers['Referer'] = referer
-        return super(SecureHTTPConnection, self).request(method, url, body,
-                                                         headers)
+        try:
+            return super(SecureHTTPConnection, self).request(method, url, body,
+                                                             headers)
+        except httplib.CannotSendRequest:
+            self.broken = True
+            raise
